from fastmask import config

from typing import Any
from rich import print, box
from rich.table import Table
from rich.console import Console
from datetime import datetime


def error_msg(msg: str, exit_: bool = True) -> None:
    print(f'[red bold]{msg}')
    if exit_:
        print("[white]Exiting...\n")
        exit()

def success_msg(msg: str) -> None:
    print(f'[green]{msg}')

def try_get(k: str, src: dict, default=None) -> Any:
    return src[k] if k in src else default

def is_none(val) -> bool:
    return val is None or len(val) == 0

def to_date(d: str) -> datetime:
    return datetime.strptime(d, config.TIME_FORMAT)


class PrettyTable():

    def __init__(self, data_: list[dict], title: str | None = None):
        self.table = Table(title=title, box=box.MINIMAL, header_style='italic')
        self.generate_cols()
        self.generate_rows(data_)

    @staticmethod
    def compute_style(datum: str, ref: dict) -> str:
        style = try_get(
            datum,
            try_get(
                'row_style',
                ref,
                {}
            ),
            ''
        )
        return f'[{style}]' if len(style) > 0 else style

    @staticmethod
    def compute_value(datum: Any, ref: dict) -> str:
        return try_get(
            'transform',
            ref,
            (lambda x: x.replace('[','\[') if x is not None else '') # escape brackets for Rich rendering
        )(datum)

    def generate_rows(self, data_: list[dict]) -> None:
        for x in data_:
            row = []
            for k,v in config.table_schema.items():
                if not try_get('hide',v,False):
                    style = self.compute_style(x[k], v)
                    value = self.compute_value(x[k], v)
                    row.append(style + value)
            self.table.add_row(*row)

    def generate_cols(self) -> None:
        for k,v in config.table_schema.items():
            if not try_get('hide',v,False):
                self.table.add_column(try_get('header',v,k), style=try_get('style',v), no_wrap=True)

    def out(self) -> None:
        console = Console()
        print()
        console.print(self.table)



from typing import Callable
import json
import requests
from datetime import datetime, timezone, timedelta


HOSTNAME = 'api.fastmail.com'
JMAP_CORE = 'urn:ietf:params:jmap:core'
MASKED_EMAIL_SCOPE = 'https://www.fastmail.com/dev/maskedemail'


class MaskedMailClient:
    """
    Client for interfacing with Fastmail's Masked Email JMAP API

    https://www.fastmail.com/developer/maskedemail/

    https://jmap.io/
    """

    def __init__(self, username: str, token: str):
        """Initialize using a username and Fastmail API token"""

        if username is not None and token is not None:
            if len(username) == 0 and len(token) == 0:
                print('No username/token found. Exiting...')
                exit()
        else:
            print('No username/token found. Exiting...')
            exit()

        self.hostname = HOSTNAME
        self.username = username
        self.token = token
        self.session = self.get_session()
        self.api_url = self.session['apiUrl']
        self.account_id = self.get_account_id()

    def get_session(self) -> dict:
        """Return the JMAP Session Resource as a Python dict

        Borrowed from Fastmail's tiny_jmap_library.py: https://github.com/fastmail/JMAP-Samples
        """

        try:
            if self.session:
                return self.session
        except:
            pass

        r = requests.get(
            "https://" + self.hostname + "/.well-known/jmap",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.token}",
            },
        )
        r.raise_for_status()
        return r.json()

    def get_api_url(self) -> str:
        """Get API URL from session"""

        if not self.session:
            self.get_session()
        return self.session['apiUrl']

    def get_account_id(self) -> str:
        """Return the accountId for the account matching self.username


        Borrowed from Fastmail's tiny_jmap_library.py: https://github.com/fastmail/JMAP-Samples
        """
        try:
            if self.account_id:
                return self.account_id
        except:
            pass

        session = self.get_session()

        return session["primaryAccounts"][JMAP_CORE]

    def __jmap_call(self, call: dict) -> dict:
        """
        Make a JMAP POST request to the API, returning the reponse as a
        Python data structure.

        Borrowed from Fastmail's tiny_jmap_library.py https://github.com/fastmail/JMAP-Samples
        """
        res = requests.post(
            self.api_url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.token}",
            },
            data=json.dumps(call),
        )
        res.raise_for_status()
        return res.json()

    def get(self, ids: list[str] | None = None, filters: Callable | None = None, sort_by: str | None = None, sort_order: str = 'asc', limit: int | None = None) -> list[dict]:
        """Get Masked Emails associated with account.

        ids [optional]: Return Masked Emails matching IDs
        filters [optional]: Optionally filter list results by field (see https://www.fastmail.com/developer/maskedemail/#object for options)
        sort_by [optional]: Field to sort results by
        sort_order [optional]: Whether to sort ascending or descending. Accepted values: 'asc' or 'desc'
        limit: Number of results to return
        """

        response = self.__jmap_call({
            'using': [JMAP_CORE, MASKED_EMAIL_SCOPE],
            'methodCalls': [
                [
                    'MaskedEmail/get',
                    {
                        'accountId': self.account_id,
                        'ids': ids,
                    },
                    'a'
                ]
            ]
        })

        masked_email_list = response['methodResponses'][0][1]['list']

        if filters is not None:
            masked_email_list = list(filter(filters, masked_email_list))

        if sort_by is not None: #TODO remove items without None in sort_by?
            masked_email_list = sorted(masked_email_list, key=lambda x: (x[sort_by] is None, x[sort_by]), reverse=(sort_order == 'desc'))

        if limit is not None:
            return masked_email_list[:limit]

        return masked_email_list

    def new(self, url: str | None = None, domain: str = '', description: str = '', state: str = 'enabled'):
        'Created a new Masked Email, optionally setting url, forDomain, description'

        response = self.__jmap_call({
            'using': [JMAP_CORE, MASKED_EMAIL_SCOPE],
            'methodCalls': [
                [
                    'MaskedEmail/set',
                    {
                        'accountId': self.account_id,
                        'create': {
                            'new-masked-email': {
                                'state': state,
                                'description': description,
                                'url': url,
                                'forDomain': domain,
                            }
                        }
                    },
                    'a'
                ]
            ]
        })
        return response['methodResponses'][0]

    def update(self, masked_id: str, changes: dict) -> dict:
        """Update an existing Masked Email"""

        response = self.__jmap_call({
            'using': [JMAP_CORE, MASKED_EMAIL_SCOPE],
            'methodCalls': [
                [
                    'MaskedEmail/set',
                    {
                        'accountId': self.account_id,
                        'update': {
                            masked_id: changes,
                        }
                    },
                    'a'
                ]
            ]
        })
        return response['methodResponses'][0]

    def get_active(self) -> list[dict]:
        """Get all active Masked Emails"""

        return self.get(filters=(
            lambda x: x['state'] == 'enabled'
        ))

    def get_disabled(self) -> list[dict]:
        """Get all blocked Masked Emails"""

        return self.get(filters=(
            lambda x: x['state'] == 'disabled'
        ))

    def get_deleted(self) -> list[dict]:
        """Get all deleted Masked Emails"""

        return self.get(filters=(
            lambda x: x['state'] == 'deleted'
        ))

    def get_unused(self) -> list[dict]:
        """Get all active Masked Emails that have not received messages"""

        return self.get(filters=(
            lambda x: x['lastMessageAt'] is None and x['state'] == 'enabled'
        ))

    def enable(self, masked_id: str) -> dict:
        """Set Masked Email to active"""

        return self.update(
            masked_id=masked_id,
            changes={
                'state': 'enabled'
            }
        )

    def disable(self, masked_id: str) -> dict:
        """Set Masked Email to blocked"""

        return self.update(
            masked_id=masked_id,
            changes={
                'state': 'disabled'
            }
        )

    def delete(self, masked_id: str) -> dict:
        """
        Delete Masked Email

        Address will appear under 'Review deleted masked address' and will remain recoverable. Does not permanently delete.
        """

        return self.update(
            masked_id=masked_id,
            changes={
                'state': 'deleted'
            }
        )

    def search(self, query: str, fields: list[str] = ['email', 'description']) -> list[dict]:
        """
        Search for Masked Emails matching query text

        Defaults to searching email and description fields only.
        """

        query = query.lower()
        return self.get(filters=(
            lambda x: any(query in x[f].lower() for f in fields if x[f] is not None)
        ))

    def get_recent(self, timeframe: timedelta = timedelta(days=3)) -> list[dict]:
        """Get recently created Masked Emails (default is 3 days)"""

        t_format = "%Y-%m-%dT%H:%M:%SZ"
        now = datetime.now(timezone.utc).replace(tzinfo=None)

        return self.get(
            filters=(
                lambda x: (now - datetime.strptime(x['createdAt'], t_format)) < timeframe
            ),
            sort_by='createdAt', sort_order='desc'
        )



from datetime import datetime

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

def to_date(d: str) -> datetime:
    return datetime.strptime(d, TIME_FORMAT)

table_schema = {
    'description': {
        'header': 'Description',
        'style': 'blue',
        'transform': (lambda x: x if x != '' else 'n/a'),
        'row_style': {
            '': 'dim',
        }
    },
    'email': {
        'header': 'Email',
        'style': 'cyan',
    },
    'id': {
        'header': 'ID',
    },
    'createdAt': {
        'header': 'Created',
        'style': 'italic',
        'transform': (lambda x: datetime.strftime(to_date(x), '%m/%d/%y'))
    },
    'lastMessageAt': {
        'header': 'Last Msg',
        'style': 'italic',
        'row_style': {
            None: 'dim',
        },
        'transform': (lambda x: datetime.strftime(to_date(x), '%m/%d/%y') if x is not None else
        'Never')
    },
    'state': {
        'header': 'State',
        'row_style': {
            'enabled': 'green',
            'disabled': 'red',
            'deleted': 'dim',
            'pending': 'yellow',
        }
    },
    'forDomain': {
        'header': 'Domain',
        'style': 'cyan bold',
        'hide': True,
    },
    'url': {
        'header': 'URL',
        'style': 'cyan',
    },
    'createdBy': {
        'header': 'By',
        'style': 'yellow',
        'transform': (lambda x: x[11:]),
        'hide': True
    },
}



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

exec(open('fastmask/version.py').read())

setup(
    name='fastmask',
    version=__version__,
    description='Python library + CLI for Fastmail masked email',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Hudson Bailey',
    author_email='dev@hudsonbailey.org',
    url='https://github.com/hdb/fastmask',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click',
        'requests',
        'python-dotenv',
        'rich',
        'pandas',
    ],
    entry_points={
        'console_scripts': [
            'fastmask = fastmask.cli:cli',
        ],
    },
)


from fastmask.masked_email import MaskedMailClient
from fastmask.utils import (
    PrettyTable, 
    error_msg, 
    success_msg,
    to_date
)

import click
from datetime import datetime, timezone, timedelta
import os
import re
import pandas as pd

skip_dotenv=os.getenv('SKIP_PYTHONDOTENV', 'False').lower() in ('true', '1', 't')
if not skip_dotenv:
    from dotenv import load_dotenv
    load_dotenv()

EMAIL_PATTERN = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'


@click.group()
@click.option(
    "--username",
    default=lambda: os.environ.get("FM_USERNAME", ""),
    show_default="FM_USERNAME"
)
@click.option(
    "--token",
    default=lambda: os.environ.get("FM_ME_TOKEN", ""),
    show_default="FM_ME_TOKEN",
)
@click.pass_context
def cli(context, username: str, token: str):
    """Manage Fastmail masked email from the command line"""
    
    context.obj = MaskedMailClient(username=username,token=token)

@cli.command(name='list')
@click.option('--limit', default=None, type=int, help='Limit number of results')
@click.option("--active", 'state', flag_value='active', help="Show only active addresses.")
@click.option("--blocked", 'state', flag_value='blocked', help="Show only blocked addresses.")
@click.option("--unused", 'state', flag_value='unused', help="Show only active + unused addresses.")
@click.option("--used", 'state', flag_value='used', help="Show only used addresses.")
@click.option("--deleted", 'state', flag_value='deleted', help="Show only deleted addresses.")
@click.option('--sort', type=click.Choice(
    ['email', 'createdAt', 'description', 'lastMessageAt', 'forDomain', 'url']
    , case_sensitive=False), help='Field to sort by')
@click.option('--desc/--asc', default=False, help='Sort order')
@click.option('--recent', default=None, type=int, help='Only show items from the past X days')
@click.option('-o', '--out', default=None, type=str, help='Output to csv')
@click.pass_obj
def list_cmd(client: MaskedMailClient, limit: int | None, state: str | None, recent: int | None, sort: str | None, desc: bool, out: str | None):
    """List masked emails associated with account"""
    
    state_map = {
        'active': (lambda x: x['state'] == 'enabled'),
        'blocked': (lambda x: x['state'] == 'disabled'),
        'deleted': (lambda x: x['state'] == 'deleted'),
        'unused': (lambda x: x['lastMessageAt'] is None and x['state'] == 'enabled'),
        'used': (lambda x: x['lastMessageAt'] is not None),
        None: (lambda x: x)
    }

    if recent is not None:
        now = datetime.now(timezone.utc).replace(tzinfo=None)
        recent_filter = (lambda x: (now - to_date(x['createdAt'])) < timedelta(days=recent))
    else:
        recent_filter = (lambda x: x)

    filters = (lambda x: state_map[state](x) and recent_filter(x))

    results = client.get(filters=filters, limit=limit, sort_by=sort, sort_order='desc' if desc else 'asc')
    if out is not None:
        pd.DataFrame(results).to_csv(out, index=False)
    else:
        PrettyTable(results, title=f'Masked Emails {client.account_id}').out()

@cli.command()
@click.argument('query', default='')
@click.option('--blank', is_flag=True)
@click.option('--field', '-f', 'fields', multiple=True, default=['email', 'description'])
@click.option('--limit', default=None, type=int, help='Limit number of results')
@click.option('-o', '--out', default=None, type=str, help='Output to csv')
@click.pass_obj
def search(client: MaskedMailClient, query: str, limit: int, fields: list[str], blank: bool, out: str | None):
    """Search for masked emails"""
    
    if blank:
        results = client.get(filters=(lambda x: x['description'] == ''), limit=limit)
    elif len(query)==0:
        error_msg('No query provided. Did you mean to use \'fastmask search --blank\'?')
    else:
        results = client.search(query=query, fields=fields)

    if out is not None:
        pd.DataFrame(results).to_csv(out, index=False)
    else:
        PrettyTable(results, title=f'Search results for "{query if query is not None else ""}"').out()

@cli.command()
@click.argument('description', default='')
@click.option('--url', type=str)
@click.option('--domain', default='', type=str)
@click.option('--pending', is_flag=True, default=False)
@click.pass_obj
def new(client: MaskedMailClient, description: str, url: str, domain: str, pending: bool):
    """Create a new masked email"""
    
    response = client.new(
        description=description,
        url=url,
        domain=domain,
        state='enabled' if not pending else 'pending'
    )

    try:
        result = [v for k,v in response[1]["created"].items()][0]
        desc_str = f'for "{description}" ' if len(description)>0 else ''
        msg = f'Successfully added email {result["email"]} {desc_str}(id: {result["id"]})'
        success_msg(msg)

    except Exception as e:
        error_msg(e)

@click.pass_obj
def update(client: MaskedMailClient, id_: str, changes: dict):
    """Update a masked email. Called by edit, activate, block, delete"""
    
    def get_masked_email_id(id_):
        if id_.startswith('masked-') and id_[7:].isnumeric():
            result = client.get(ids=[id_])
        elif id_.isnumeric():
            result = client.get(ids=[f'masked-{id_}'])
        elif re.match(EMAIL_PATTERN, id_):
            result = client.get(filters=(lambda x: x['email'] == id_))
        else:
            result = client.get(filters=(lambda x: x['description'].lower() == id_))

        if len(result)>0:
            return result[0]['id']
        else:
            error_msg(f'Unable to find Masked Email "{id_}"')

    masked_email_id = get_masked_email_id(id_)

    changes = {k:v for k,v in changes.items() if v is not None}

    if len(changes)<1:
        error_msg('Specify fields to update')

    response = client.update(masked_id=masked_email_id, changes=changes)

    return response

@cli.command()
@click.argument('id_', required=True)
@click.option('--description', type=str)
@click.option('--url', type=str)
@click.option('--domain', type=str)
@click.pass_obj
def edit(client: MaskedMailClient, id_: str, description: str, url: str | None, domain: str | None):
    """Edit information associated with a masked email"""
    
    changes = {
        'description': description,
        'url': url,
        'forDomain': domain,
    }

    response = update(id_=id_, changes=changes)

    try:
        msg = f'Successfully edited email {[k for k,v in response[1]["updated"].items()][0]}'
        success_msg(msg)

    except Exception as e:
        error_msg(e)

@cli.command()
@click.argument('id_', required=True)
@click.pass_obj
def activate(client: MaskedMailClient, id_: str):
    """Set state of masked email to Active"""

    response = update(id_=id_, changes={'state': 'enabled'})
    try:
        msg = f'Successfully activated email {[k for k,v in response[1]["updated"].items()][0]}'
        success_msg(msg)

    except Exception as e:
        error_msg(e)

@cli.command()
@click.argument('id_', required=True)
@click.pass_obj
def block(client: MaskedMailClient, id_: str):
    """Set state of masked email to Blocked"""

    response = update(id_=id_, changes={'state': 'disabled'})

    try:
        msg = f'Successfully blocked email {[k for k,v in response[1]["updated"].items()][0]}'
        success_msg(msg)

    except Exception as e:
        error_msg(e)

@cli.command()
@click.argument('id_', required=True)
@click.pass_obj
def delete(client: MaskedMailClient, id_: str):
    """Delete masked email"""

    response = update(id_=id_, changes={'state': 'deleted'})

    try:
        msg = f'Successfully deleted email {[k for k,v in response[1]["updated"].items()][0]}'
        success_msg(msg)

    except Exception as e:
        error_msg(e)
